<?php
/**
 * Implements _webform_defaults_component().
 */
function _webform_defaults_address() {
  return array(
    'name' => '',
    'form_key' => NULL,
    'required' => FALSE,
    'pid' => 0,
    'weight' => 0,
    'extra' => array(
      'po_box' => FALSE,
      'addresscheck' => 'none',
      'city_options' => 'textfield',
      'cities_piped_list' => '',
      'placeholders' => FALSE,
      'private' => FALSE,
    ),
  );
}

/**
 * Implements _webform_edit_component().
 */
function _webform_edit_address($component) {
  $form = array();

  $form['extra']['po_box'] = array(
    '#type' => 'checkbox',
    '#title' => t('PO Box'),
    '#description' => t('Check this if this is a PO box address. Street name, number and addition will dissapear and PO box field becomes available. NOTE: address checks are not available with a PO box.'),
    '#default_value' => $component['extra']['po_box'],
  );
  $addresscheck_options = array();
  $addresscheck_options['none'] = t('None - no checks are performed');
  $addresscheck_options['postcodenl'] = t('Postcode.nl - only addresses in the Netherlands are allowed');
  if (function_exists('dvg_stuf_bg_adresgegevens')) {
    $addresscheck_options['stuf-bg'] = t('StUF-BG Addresscheck - only addresses in municipality are allowed');
  }
  $form['extra']['addresscheck'] = array(
    '#type' => 'select',
    '#title' => t('Address check'),
    '#description' => t('Optional. Select the service to complete streetname and city based on postal code + house number.'),
    '#options' => $addresscheck_options,
    '#default_value' => $component['extra']['addresscheck'],
  );
  $form['extra']['city_options'] = array(
    '#type' => 'select',
    '#title' => t('City options'),
    '#description' => t('Do you want to display the city field as a select list or free textfield? NOTE: address checks are not available if predefined list is used.'),
    '#options' => array(
      'textfield' => t('Textfield'),
      'select' => t('Predefined list'),
    ),
    '#default_value' => $component['extra']['city_options'],
  );
  $form['extra']['cities_piped_list'] = array(
    '#type' => 'textarea',
    '#title' => t('Predefined list of cities'),
    '#description' => t('If you choose predefined list above, please provide a piped list (key|value) of all available cities.'),
    '#default_value' => $component['extra']['cities_piped_list'],
  );
  $form['extra']['placeholders'] = array(
    '#type' => 'checkbox',
    '#title' => t('Placeholders'),
    '#description' => t('Use a placeholder instead of the label. The placeholder will be shown in the field until the user starts entering a value.'),
    '#default_value' => $component['extra']['placeholders'],
  );

  return $form;
}

/**
 * Implements _webform_render_component().
 */
function _webform_render_address($component, $value = NULL, $filter = TRUE) {
  // Wrap the form in a container, in order to correctly apply the weight.
  $elements = array(
    '#type' => 'container',
    '#attached' => array(
      'css' => array(
        drupal_get_path('module', 'dvg_webform_components') . '/css/address.css',
      ),
      'js' => array(
        drupal_get_path('module', 'dvg_webform_components') . '/js/address.js',
      ),
    ),
    '#weight' => $component['weight'],
    '#theme_wrappers' => array('webform_element'),
    '#webform_component' => $component,
    '#element_validate' => array('dvg_webform_components_address_validate_addresscheck'),
  );
  $elements['postal_code'] = array(
    '#type' => 'textfield',
    '#title' => t('Postal code'),
    '#title_display' => $component['extra']['title_display'] ? $component['extra']['title_display'] : 'before',
    '#default_value' => isset($value['postal_code']) ? $value['postal_code'] : '',
    '#required' => $component['required'],
    '#attributes' => array(),
    '#size' => 12,
    '#theme_wrappers' => array('webform_element'),
    '#webform_component' => array('type' => $component['type'] . '_postal_code') + $component,
    '#element_validate' => array('dvg_webform_components_address_validate'),
  );
  if (empty($component['extra']['po_box'])) {
    $elements['house_number'] = array(
      '#type' => 'textfield',
      '#title' => t('House number'),
      '#title_display' => $component['extra']['title_display'] ? $component['extra']['title_display'] : 'before',
      '#default_value' => isset($value['house_number']) ? $value['house_number'] : '',
      '#required' => $component['required'],
      '#attributes' => array(),
      '#size' => 12,
      '#theme_wrappers' => array('webform_element'),
      '#webform_component' => array('type' => $component['type'] . '_house_number') + $component,
      '#element_validate' => array('dvg_webform_components_address_validate'),
    );
    $elements['house_number_addition'] = array(
      '#type' => 'textfield',
      '#title' => t('Number addition'),
      '#title_display' => $component['extra']['title_display'] ? $component['extra']['title_display'] : 'before',
      '#default_value' => isset($value['house_number']) ? $value['house_number'] : '',
      '#required' => FALSE,
      '#attributes' => array(),
      '#size' => 12,
      '#theme_wrappers' => array('webform_element'),
      '#webform_component' => array('type' => $component['type'] . '_house_number_addition') + $component,
      '#element_validate' => array('dvg_webform_components_address_validate'),
    );
    $elements['street_name'] = array(
      '#type' => 'textfield',
      '#title' => t('Street name'),
      '#title_display' => $component['extra']['title_display'] ? $component['extra']['title_display'] : 'before',
      '#default_value' => isset($value['street_name']) ? $value['street_name'] : '',
      '#required' => $component['required'],
      '#attributes' => array(),
      '#theme_wrappers' => array('webform_element'),
      '#webform_component' => array('type' => $component['type'] . '_street_name') + $component,
    );
  } else {
    $elements['po_box'] = array(
      '#type' => 'textfield',
      '#title' => t('PO Box'),
      '#title_display' => $component['extra']['title_display'] ? $component['extra']['title_display'] : 'before',
      '#default_value' => isset($value['po_box']) ? $value['po_box'] : '',
      '#required' => $component['required'],
      '#attributes' => array(),
      '#size' => 12,
      '#theme_wrappers' => array('webform_element'),
      '#webform_component' => array('type' => $component['type'] . '_po_box') + $component,
      '#element_validate' => array('dvg_webform_components_address_validate'),
    );
  }
  if ($component['extra']['city_options'] == 'textfield') {
    $elements['city'] = array(
      '#type' => 'textfield',
      '#title' => t('City'),
      '#title_display' => $component['extra']['title_display'] ? $component['extra']['title_display'] : 'before',
      '#default_value' => isset($value['city']) ? $value['city'] : '',
      '#required' => $component['required'],
      '#attributes' => array(),
      '#theme_wrappers' => array('webform_element'),
      '#webform_component' => array('type' => $component['type'] . '_city') + $component,
    );
  }

  if ($component['extra']['city_options'] == 'select') {
    $options = list_extract_allowed_values($component['extra']['cities_piped_list'], 'list_text', FALSE);
    $elements['city'] = array(
      '#type' => 'select',
      '#title' => t('City'),
      '#options' => $options,
      '#title_display' => $component['extra']['title_display'] ? $component['extra']['title_display'] : 'before',
      '#default_value' => isset($value['city']) ? $value['city'] : '',
      '#required' => $component['required'],
      '#attributes' => array(),
      '#theme_wrappers' => array('webform_element'),
      '#webform_component' => array('type' => $component['type'] . '_city') + $component,
    );
  }

  if ($component['extra']['addresscheck'] == 'stuf-bg' && empty(variable_get('dvg_stuf_bg_aoa_api_url'))) {
    watchdog('dvg_webform_components', 'Addresscheck set to stuf-bg but variable dvg_stuf_bg_aoa_api_url not set ', array(), WATCHDOG_WARNING, $link = NULL);
  }
  if ($component['extra']['addresscheck'] == 'postcodenl' && empty(variable_get('dvg_postcodenl_key'))) {
    watchdog('dvg_webform_components', 'Addresscheck set to postcode.nl but variable dvg_postcodenl_key not set ', array(), WATCHDOG_WARNING, $link = NULL);
  }
  if ($component['extra']['addresscheck'] == 'postcodenl' && empty(variable_get('dvg_postcodenl_secret'))) {
    watchdog('dvg_webform_components', 'Addresscheck set to postcode.nl but variable dvg_postcodenl_secret not set ', array(), WATCHDOG_WARNING, $link = NULL);
  }

  $ajax = FALSE;
  if ($component['extra']['addresscheck'] != 'none' && $component['extra']['city_options'] != 'select' && !$component['extra']['po_box']) {
    $ajax = TRUE;
  }

  // ajax calls for addresschecks only added if :
  // - selected service
  // - cities not a list
  // - not a po box
  if ($ajax) {
    $elements['postal_code']['#ajax'] = array(
      'callback' => '_address_ajax_callback',
    );
    $elements['house_number']['#ajax'] = array(
      'callback' => '_address_ajax_callback',
    );
    $elements['street_name']['#attributes'] = array(
      'class' => array(drupal_html_class('ajax-callback-street-name--' . $component['form_key'])),
    );
    $elements['city']['#attributes'] = array(
      'class' => array(drupal_html_class('ajax-callback-city--' . $component['form_key'])),
    );
  }

  // Set the label as a place holder.
  if ($component['extra']['placeholders']) {

    $elements['po_box']['#title_display'] = 'invisible';
    $elements['po_box']['#attributes']['placeholder'] = t('PO Box');

    $elements['postal_code']['#title_display'] = 'invisible';
    $elements['postal_code']['#attributes']['placeholder'] = t('Postal code');

    $elements['house_number']['#title_display'] = 'invisible';
    $elements['house_number']['#attributes']['placeholder'] = t('House number');

    $elements['house_number_addition']['#title_display'] = 'invisible';
    $elements['house_number_addition']['#attributes']['placeholder'] = t('Addition');

    $elements['street_name']['#title_display'] = 'invisible';
    $elements['street_name']['#attributes']['placeholder'] = t('Street name');

    $elements['city']['#title_display'] = 'invisible';
    $elements['city']['#attributes']['placeholder'] = t('City');
  }

  return $elements;
}

/**
 * Retrieve street name and city belonging to postal code and house number.
 *
 * Callback for _webform_render_address_auto().
 *
 * NB: Normally, this function should be kept as simple as possible (i.e., just
 *     return the portion of the form that should be adjusted). However, since
 *     Webform doesn't provide the $form_state in a simple manner and we only
 *     provide additional functionallity to help the visitor, we allow us to do
 *     a little bit more in this function.. (otherwise we have to use the
 *     hook_form_alter() function to add AJAX components).
 */
function _address_ajax_callback($form, $form_state) {
  // To authenticate to the postcode.nl API, we need the login information.

  $service = $form_state['triggering_element']['#webform_component']['extra']['addresscheck'];

  // Get the key used for this Webform component.
  $form_key = $form_state['triggering_element']['#webform_component']['form_key'];

  // Get the provided postal code and house number.
  $values = $form_state['values'];
  foreach ($form_state['triggering_element']['#parents'] as $parent) {
    $values = $values[$parent];

    if ($parent == $form_key) {
      break;
    }
  }
  $postal_code = $values['postal_code'];
  $house_number = $values['house_number'];

  // Perform the lookup using the service.
  $data = _address_autofill($postal_code, $house_number, $service);

  // React on the provided information.
  if (!empty($data['street']) && !empty($data['city'])) {
    $commands = array();
    $arguments = array(
      'formKey' => $form_key,
      'street' => check_plain($data['street']),
      'city' => check_plain($data['city']),
    );
    $arguments = drupal_json_encode($arguments);
    $commands[] = ajax_command_invoke(NULL, 'webformAddressInject', array($arguments));

    return array(
      '#type' => 'ajax',
      '#commands' => $commands,
    );
  }
}

/**
 * Wrapper function which calls different services based on settings.
 *
 * @param string $service
 *   Type of service: currently: postcodenl, aoa (dvg-stuf-bg)
 * @param string $postal_code
 *   The postal code of the address.
 * @param string $house_number
 *   The house number of the address.
 */
function _address_autofill($postal_code, $house_number, $service) {

  $data = array();

  switch($service) {
    case 'stuf-bg':
      $prefill = dvg_stuf_bg_adresgegevens($postal_code, $house_number, $house_letter = '', $house_number_addition = '', $reset = FALSE);
      if ($prefill) {
        $data['street'] = $prefill['gor.straatnaam']->_;
        $data['city'] = $prefill['wpl.woonplaatsNaam']->_;
      }
      break;
    case 'postcodenl':
      $key = variable_get('dvg_postcodenl_key');
      $secret = variable_get('dvg_postcodenl_secret');
      $data = _address_postcode_nl_lookup($key, $secret, $postal_code, $house_number);
      break;
  }
  return $data;
}

/**
 * Retrieve street name and city belonging to postal code and house number.
 *
 * This function makes a call to the postcode.nl API to get the street name and
 * city that is associated with the postal code and house number.
 *
 * @param string $key
 *   The key that should be used to identify at the postcode.nl API service.
 * @param string $secret
 *   The secret that belongs to the $key.
 * @param string $postal_code
 *   The postal code of the address.
 * @param string $house_number
 *   The house number of the address.
 *
 * @return string[]
 *   The name of the street can be found at the index 'street' and the name of
 *   the name of the city at the index 'city' in the array.
 */
function _address_postcode_nl_lookup($key, $secret, $postal_code, $house_number) {
  // First, parse the postal code and house number (which should split the
  // addition).
  $postal_code = _address_parse_postal_code($postal_code);
  list($house_number, $house_number_addition) = _address_parse_house_number($house_number);

  // Both the postal code and the house number should have a value.
  if (empty($postal_code) || empty($house_number)) {
    return array();
  }

  // We prevent making unnecessary lookups by caching results, also wrong
  // lookups!
  $cache_key = 'dvg_address:' . $postal_code . ':' . $house_number . ':' . $house_number_addition;

  // Use the local memory as the first level cache.
  $data = &drupal_static(__FUNCTION__);
  if (!isset($data)) {
    // When not stored in local memory, check the cache stored in the database.
    if ($cache = cache_get($cache_key)) {
      $data = $cache->data;
    }
    else {
      // Construct the URL to perform the lookup with.
      $url = 'https://' . $key . ':' . $secret . '@api.postcode.nl/rest/addresses/' . $postal_code . '/' . $house_number . '/' . $house_number_addition;

      $response = drupal_http_request(url($url));

      // When the lookup went correctly, parse the JSON data.
      if ($response->code == 200) {
        $data = drupal_json_decode($response->data);
      }

      // Store the result in the cache of the database.
      //
      // NB: if the lookup didn't responded in a correct result, we still store
      // a 'mark' in the database. In this way also invalid lookups will be
      // served from cache the next time.
      cache_set($cache_key, $data, 'cache', time() + 900);
    }
  }

  return $data;
}

/**
 * Parse a postal code to be used in the request to postcode.nl API.
 *
 * @param string $postal_code
 *   The postal code that should be normalized.
 *
 * @return string
 *   The normalized postal code.
 */
function _address_parse_postal_code($postal_code) {
  // First, remove any whitespaces.
  $postal_code = str_replace(' ', '', trim($postal_code));

  // If the postal code is in the form of 1234AB, than return the trimmed
  // version.
  if ((boolean) preg_match('~^\d{4}[[:alpha:]]{2}$~', $postal_code)) {
    return $postal_code;
  }

  return '';
}

/**
 * Parse the house number by separating the number from the additional value.
 *
 * @param string $house_number_full
 *   The house number of the address (including an possible addition).
 *
 * @return string[]
 *   Normalized version of the house number, where the first index of the array
 *   contains the number and the second value is the addition.
 */
function _address_parse_house_number($house_number_full) {
  // Remove white spaces before and after the provided house number (including
  // addition).
  $house_number_full = trim($house_number_full);

  $house_number = '';
  $house_number_addition = '';
  if (preg_match('~^(?P<house_number>\d+)[^[:alnum:]]*(?P<addition>[[:alnum:]]*)$~', $house_number_full, $match)) {
    $house_number = $match['house_number'];
    $house_number_addition = !empty($match['addition']) ? $match['addition'] : '';
  }

  return array($house_number, $house_number_addition);
}

/**
 * Implements _webform_display_component().
 */
function _webform_display_address($component, $value, $format = 'html') {
  return array(
    '#title' => $component['name'],
    '#weight' => $component['weight'],
    '#theme' => 'webform_display_address_auto',
    '#theme_wrappers' => $format == 'html' ? array('webform_element') : array('webform_element_text'),
    '#post_render' => array('webform_element_wrapper'),
    '#component' => $component,
    '#format' => $format,
    '#value' => isset($value) ? oaddress_display_format($value) : '',
  );
}

/**
 * Implements _webform_theme_component().
 */
function _webform_theme_address() {
  return array(
    'webform_display_address' => array(
      'render element' => 'element',
    ),
  );
}

/**
 * Returns HTML for displaying the address.
 *
 * @param array $variables
 *   An associative array containing:
 *   - element: An associative array containing the properties of the element.
 *     Properties used: #format, #value.
 *
 * @see theme_webform_element
 *
 * @ingroup themeable
 */
function theme_webform_display_address(array $variables) {
  $element = $variables['element'];
  $value = $element['#format'] == 'html' ? check_plain($element['#value']) : $element['#value'];

  return $value !== '' ? $value : ' ';
}

/**
 * Implements _webform_table_component().
 */
function _webform_table_address($component, $value) {
  return check_plain($value['street_name'] . ' ' . $value['house_number']) . '<br />' .
  check_plain($value['postal_code'] . ' ' . $value['city']);
}

/**
 * Implements _webform_csv_headers_component().
 */
function _webform_csv_headers_address($component, $export_options) {
  $header = array();
  $header[0] = array('');
  $header[1] = array($component['name']);
  $header[2] = array();

  $header[2][] = t('Street name');
  $header[2][] = t('House number');
  $header[2][] = t('Postal code');
  $header[2][] = t('City');

  return $header;
}

/**
 * Implements _webform_csv_data_component().
 */
function _webform_csv_data_address($component, $export_options, $value) {
  $return = array();

  $return[] = $value['street_name'];
  $return[] = $value['house_number'];
  $return[] = $value['postal_code'];
  $return[] = $value['city'];

  return $return;
}

/**
 * Format the address in a textual representation.
 *
 * @param string[] $value
 *   An associative array containing the address information (i.e., street name,
 *   house number, postal code and the city).
 *
 * @return string
 *   A formatted string representing the address.
 */
function address_display_format(array $value) {
  return $value['street_name'] . ' ' . $value['house_number'] . ', ' . $value['postal_code'] . ' ' . $value['city'];
}

/**
 * Validate entire form, need
 */
function dvg_webform_components_address_validate_addresscheck($element, &$form_state) {
  $postal_code = $element['postal_code']['#value'];
  $house_number = $element['house_number']['#value'];
  $service = $element['#webform_component']['extra']['addresscheck'];

  $city = $element['city']['#value'];
  $street_name = $element['street_name']['#value'];

  // perform the lookup once again
  $data = _address_autofill($postal_code, $house_number, $service);

  // React on the provided information.
  if (empty($data) || (!empty($data['street']) && !empty($data['city'])) || $data['street'] != $street_name || $data['city'] != $city) {
    $error = t('This address is not available in %site.', array('%site' => variable_get('site_name', 'this site')));
    form_set_error($element['#webform_component']['type'], $error);
  }
}

/**
 * Validate seperate elements
 */
function dvg_webform_components_address_validate($element, &$form_state) {

  switch($element['#webform_component']['type']) {
    case 'address_postal_code':
      $result = _address_validate_postal_code($element['#value']);
      break;
    case 'address_house_number':
      $result = _address_validate_house_number($element['#value']);
      break;
    case 'address_house_number_addition':
      $result = _address_validate_house_number_addition($element['#value']);
      break;
    case 'address_po_box':
      $result = _address_validate_po_box($element['#value']);
      break;
  }

  if (isset($result['error'])) {
    form_set_error($element['#webform_component']['type'], $result['error']);
  }
}

/**
 * Postal code validation
 */
function _address_validate_postal_code($postal_code) {
  $return = array('country' => 'NL');
  if (!empty($postal_code)) {
    if (preg_match('/^([1-9][0-9]{3}) ?([A-Za-z]{2})$/', $postal_code, $matches) && !in_array($matches[2], array(
        'SS',
        'SD',
        'SA'
      ), TRUE)
    ) {
      $return['postal_code'] = $matches[1] . ' ' . $matches[2];
    }
    else {
      $return['error'] = t('Invalid postal code. Postal codes in the Netherlands are like "9999 AA". They never start with zero and the letters are never "SS", "SD", or "SA".');
    }
  }
  return $return;
}

/**
 * Validation handler for house number.
 */
function _address_validate_house_number($house_number) {
  $return = array();
  $value = trim($house_number);
  if (!empty($value) && !preg_match('/^[1-9]{1}[0-9]{0,4}$/i', $value)) {
    $return['error'] = t('The house number has to be an integer in the range 1 till 99999');
  }
  return $return;
}

/**
 * Validation handler for house number remark.
 */
function _address_validate_house_number_addition($house_number_addition) {
  $return = array();
  $value = trim($house_number_addition);
  if (!empty($value) && !preg_match('/^[a-z1-4]{0,4}$/i', $value)) {
    $return['error'] = t('The house number remark is max 4 characters and exists of the letters (A till Z) and/or digits (0 till 9)');
  }
  return $return;
}

/**
 * Validation handler for house number remark.
 */
function _address_validate_po_box($po_box) {
  $return = array();
  $value = trim($po_box);
  if (!empty($value) && !preg_match('/^[1-9]{1}[0-9]{0,4}$/i', $value)) {
    $return['error'] = t('The PO box has to be an integer in the range 1 till 99999');
  }
  return $return;
}